require 'kafka_rest_client/configuration'
require 'kafka_rest_client/errors'
require 'kafka_rest_client/utils'
require 'httparty'

module KafkaRestClient
  class AvroProducer
    AVRO_AS_JSON = 'application/vnd.kafka.avro.v1+json'.freeze

    attr_reader :config
    private :config

    def initialize(config = Configuration.current)
      @config = config
    end

    def enabled?
      if config.kafka_proxy_url && config.schema_registry_url
        true
      else
        false
      end
    end

    def produce(topic, events, options = {})
      if enabled?
        events = [events].flatten.map { |event| event_to_hash event }
        if options[:annotate_optional_fields]
          serialized_events = annotate_events_optional_fields(topic, events, options)
        end
        payload  = build_event_payload(topic, (serialized_events || events), options)
        response = post_event_to_kafka(topic, payload)
        config.logger.debug("Produced to Kafka topic #{topic}: #{payload.inspect}")
        response
      end
    end

    private

    def event_to_hash(event)
      if event.respond_to? :as_json
        event.as_json
      else
        Utils.deep_stringify_keys event
      end
    end

    def annotate_optional_fields(event, fields)
      event.each do |k, v|
        if v.is_a?(Hash)
          fail NestedHashSerializationNotImplemented
        elsif v && fields[k]
          event[k] = { fields[k]['type'] => event[k] }
        end
      end
    end

    def select_schema(topic, options)
      if options[:value_schema]
        options[:value_schema]
      elsif options[:value_schema_id]
        get_schema_by_id(options[:value_schema_id]).fetch('schema')
      else
        get_latest_schema(topic).fetch('schema')
      end
    end

    # Transforms an optional schema attribute
    # From {"name"=>"temperature", "type"=>["null", "double"]}
    # To {"temperature"=>{"type"=>"double"}}
    def schema_optional_fields(raw_schema)
      JSON.parse(raw_schema).fetch('fields').inject({}) do |hash, attribute|
        if attribute['type'].is_a?(Array) && attribute['type'].include?('null')
          hash[attribute['name']] = { 'type' => (attribute['type'] - ['null']).first }
        end
        hash
      end
    end

    def annotate_events_optional_fields(topic, events, options)
      raw_schema = select_schema(topic, options)
      optional_fields = schema_optional_fields(raw_schema)
      events.map { |event| annotate_optional_fields(event, optional_fields) }
    end

    def build_records(events, options)
      key = options[:key].to_s

      events.map { |event|
        if key == ""
          { value: event }
        else
          key_hash = Zlib::crc32(event[key].to_s) # Hacky solution for getting an int hash due to https://github.com/confluentinc/kafka-rest/issues/118
          { value: event, key: key_hash }
        end
      }
    end

    def build_event_payload(topic, events, options)
      payload = { records: build_records(events, options) }.merge(
        options.select { |k, _| [:value_schema_id, :value_schema].include?(k) }
      )

      if !options[:value_schema_id] && !options[:value_schema]
        payload[:value_schema_id] = get_latest_schema(topic).fetch('id')
      end

      if !options[:key].nil?
        payload[:key_schema] = { type: "long" }.to_json # Should be string, workaround for https://github.com/confluentinc/kafka-rest/issues/118
      end

      payload
    end

    def get_schema(id, url)
      response = HTTParty.get(url, timeout: config.timeout)

      if response.code == 200
        JSON.parse(response.body)
      elsif response.code == 404
        fail SchemaNotFoundError, "Schema for #{id} not found: #{response.body}"
      elsif response.code == 500
        fail InternalServerError, "Schema registry internal error: #{response.body}"
      elsif response.code == 503
        fail BadGateway, "Bad gateway: #{response.body}"
      else
        fail Error, "HTTP status: #{response.code} #{response.body}"
      end
    end

    def get_schema_by_id(id)
      url = "#{config.schema_registry_url}/schemas/ids/#{id}"
      get_schema(id, url)
    end

    def get_latest_schema(topic)
      url = "#{config.schema_registry_url}/subjects/#{topic}-value/versions/latest"
      get_schema('latest', url)
    end

    def post_event_to_kafka(topic, message)
      response = HTTParty.post(
        "#{config.kafka_proxy_url}/topics/#{topic}",
        headers: { 'Content-Type' => AVRO_AS_JSON },
        body: message.to_json,
        timeout: config.timeout
      )

      if response.code == 200
        config.logger.debug("#{message[:records].count} event(s) published on #{topic}")
        JSON.parse(response.body)
      elsif response.code == 404
        fail TopicNotFoundError,
             "Topic #{topic} not found in schema registry"
      elsif response.code == 422
        fail SchemaValidationError,
             "Message #{message} does not match schema for #{topic} or is otherwise invalid"
      elsif response.code == 500
        fail InternalServerError,
             'Kafka Rest Proxy internal error (check zookeeper, kafka services)'
      elsif response.code == 503
        fail BadGateway, "Bad gateway: #{response.body}"
      else
        fail Error, "HTTP status: #{response.code} #{response.body}"
      end
    end
  end
end
